try:
    import nlopt
    using_nlopt = True
except ImportError:
    using_nlopt = False

if using_nlopt:

    NLOPTIMIZERS = {'gn_direct' :nlopt.GN_DIRECT,
                    'gn_ags':nlopt.GN_AGS,
                    'gn_esch':nlopt.GN_ESCH,
                    'gn_isres':nlopt.GN_ISRES,
                    'gn_crs2lm':nlopt.GN_CRS2_LM,
                    'gn_directl':nlopt.GN_DIRECT_L,
                    'gn_directr':nlopt.GN_DIRECT_L_RAND,
                    'gn_directo':nlopt.GN_ORIG_DIRECT}


    def nlopt_cube_factory(objective ,n_trials, n_dim, with_count, method='gn_direct'):
        optim = NLOPTIMIZERS[method]

        global feval_count
        feval_count = 0

        lb = [0. for _ in range(n_dim)]
        ub = [1. for _ in range(n_dim)]

        def _objective(u, grad) -> float:
            global feval_count
            feval_count += 1
            return objective(u)

        opt = nlopt.opt(optim, n_dim)
        opt.set_lower_bounds(lb)
        opt.set_upper_bounds(ub)
        opt.set_min_objective(_objective)
        opt.set_maxeval(n_trials - 1)  # Groan
        best_x = opt.optimize([0.5] * n_dim)
        best_val = _objective(best_x, grad=None)  # <-- Stupid

        return (best_val, best_x, feval_count) if with_count else (best_val, best_x)


    def nlopt_direct_cube(objective ,n_trials, n_dim, with_count):
        return nlopt_cube_factory(objective=objective,n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='gn_direct')


    def nlopt_ags_cube(objective ,n_trials, n_dim, with_count):
        return nlopt_cube_factory(objective=objective,n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='gn_ags')


    def nlopt_esch_cube(objective ,n_trials, n_dim, with_count):
        return nlopt_cube_factory(objective=objective,n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='gn_esch')


    def nlopt_isres_cube(objective ,n_trials, n_dim, with_count):
        return nlopt_cube_factory(objective=objective,n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='gn_isres')


    def nlopt_mlsl_cube(objective ,n_trials, n_dim, with_count):
        return nlopt_cube_factory(objective=objective,n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='gn_mlsl')


    def nlopt_crs2lm_cube(objective ,n_trials, n_dim, with_count):
        return nlopt_cube_factory(objective=objective,n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='gn_crs2lm')


    def nlopt_directl_cube(objective ,n_trials, n_dim, with_count):
        return nlopt_cube_factory(objective=objective,n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='gn_directl')


    def nlopt_directr_cube(objective ,n_trials, n_dim, with_count):
        return nlopt_cube_factory(objective=objective,n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='gn_directr')


    def nlopt_directo_cube(objective ,n_trials, n_dim, with_count):
        return nlopt_cube_factory(objective=objective,n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='gn_directo')


    # SEEMINGLY_BROKEN = [nlopt_mlsl_cube] # Will crash Python with uncaught exception


    NLOPT_OPTIMIZERS = [nlopt_direct_cube, nlopt_ags_cube, nlopt_esch_cube,
                        nlopt_isres_cube, nlopt_crs2lm_cube, nlopt_directr_cube,
                        nlopt_directo_cube, nlopt_directl_cube]
    NLOPT_TOP_OPTIMIZERS = [nlopt_directr_cube]
else:
    NLOPT_OPTIMIZERS = []
    NLOPT_TOP_OPTIMIZERS = []


if __name__=='__main__':
    assert using_nlopt
    from humpday.objectives.classic import CLASSIC_OBJECTIVES
    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in NLOPT_OPTIMIZERS:
            print((optimizer(objective, n_trials=250, n_dim=6, with_count=True)))


try:
    import optuna
    using_optuna = True
except ImportError:
    using_optuna = False

if using_optuna:

    import optuna
    from optuna.logging import CRITICAL
    from humpday.objectives.classic import CLASSIC_OBJECTIVES
    import logging
    from humpday.transforms.zcurves import curl_factory

    logging.getLogger('optuna').setLevel(logging.ERROR)


    def optuna_cube_factory(objective, n_trials, n_dim, with_count=False, method=None):

        if method.lower()=='random':
            sampler = optuna.samplers.RandomSampler()
        elif method.lower()=='cmaes':
            sampler = optuna.samplers.CmaEsSampler()
        elif method.lower()=='tpe':
            sampler = optuna.samplers.TPESampler()
        else:
            raise ValueError('random, cmaes, tpe or grid please')

        global feval_count
        feval_count = 0

        def cube_objective(trial):
            global feval_count
            us = [ trial.suggest_float('u'+str(i),0,1) for i in range(n_dim)]
            feval_count += 1
            return objective(us)

        optuna.logging.set_verbosity(CRITICAL)
        study = optuna.create_study(sampler=sampler)
        study.optimize(cube_objective,n_trials=n_trials)

        best_x = [ study.best_params['u'+str(i)] for i in range(n_dim) ]

        if with_count:
            return study.best_value, best_x, feval_count
        else:
            return study.best_value, best_x


    def optuna_random_cube(objective, n_trials,n_dim, with_count=False):
        return optuna_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='random')


    def optuna_random_cube_clone(objective, n_trials,n_dim, with_count=False): # duplicate
        return optuna_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='random')


    def optuna_random_cube_clone_1(objective, n_trials,n_dim, with_count=False): # duplicate
        return optuna_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='random')


    def optuna_random_cube_clone_2(objective, n_trials,n_dim, with_count=False): # duplicate
        return optuna_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='random')


    def optuna_cmaes_cube(objective, n_trials, n_dim, with_count=False):
        return optuna_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='cmaes')


    def optuna_tpe_cube(objective, n_trials, n_dim, with_count=False):
        return optuna_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='tpe')


    def optuna_cmaes_curl2_cube(objective, n_trials, n_dim, with_count=False):
        # Highly experimental
        return curl_factory(optimizer=optuna_cmaes_cube, objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,d=2)


    def optuna_cmaes_curl3_cube(objective, n_trials, n_dim, with_count=False):
        # Highly experimental
        return curl_factory(optimizer=optuna_cmaes_cube, objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,d=3)


    OPTUNA_OPTIMIZERS = [optuna_cmaes_cube,
                         optuna_tpe_cube, optuna_random_cube, optuna_random_cube_clone, optuna_random_cube_clone_1,
                         optuna_random_cube_clone_2]
    OPTUNA_TOP_OPTIMIZERS = [optuna_cmaes_cube]
else:
    OPTUNA_OPTIMIZERS = []
    OPTUNA_TOP_OPTIMIZERS = []



if __name__=='__main__':
    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in OPTUNA_OPTIMIZERS:
            print((optimizer(objective, n_trials=250, n_dim=16, with_count=True)))



from humpday.optimizers.alloptimizers import OPTIMIZERS
from humpday.objectives.allobjectives import OBJECTIVES
from humpday.objectives.allobjectives import CLASSIC_OBJECTIVES
from humpday.comparison.suggestions import suggest, recommend
from humpday.optimizers.meta import minimize
from humpday.comparison.odious import compare, points_race





# The horse racing problem
import numpy as np
import math
from winning.std_calibration import std_ability_implied_dividends

global ABILITIES
ABILITIES = None
HORSE_DIM = 500 # Maximum dimension
global DIVIDENDS
DIVIDENDS = None



def make_abilities():
    global ABILITIES
    if ABILITIES is None:
        from datetime import datetime
        day_of_year = datetime.now().timetuple().tm_yday
        np.random.seed(day_of_year)
        ABILITIES = sorted(np.random.randn(HORSE_DIM))
    return ABILITIES


def make_dividends(n_dim):
    global DIVIDENDS
    if DIVIDENDS is None:
        DIVIDENDS = dict()
    if not DIVIDENDS.get(n_dim):
        DIVIDENDS[n_dim] = std_ability_implied_dividends(ability=make_abilities()[:n_dim])
    return DIVIDENDS[n_dim]


def cube_to_ability(u:[float])->[float]:
    offsets =  [0] + list(np.arctanh(np.array(np.minimum(u, 1 - 1e-5))))
    ability = [ min(5,o/100) for o in offsets ]
    return ability


def horse_dividends_on_cube(u:[float])->float:
    """ Find relative ability matching market prices
    :param u:  Determines ability of 2nd through last horse, via arctanh
    :return: [float] of dividends (return on betting $1)
    """
    n_dim = len(u)+1
    dividends = make_dividends(n_dim)
    ability = cube_to_ability(u=u)
    implied_dividends = std_ability_implied_dividends(ability=ability)
    discrepancy = np.mean( [ abs(math.sqrt(d1)-math.sqrt(d2)) for d1,d2 in zip(dividends,implied_dividends)])
    return discrepancy


HORSE_OBJECTIVES = [ horse_dividends_on_cube ] # Seems unstable and needs fixin'



if __name__=='__main__':
    from humpday.optimizers.nevergradcube import nevergrad_ngopt8_cube
    v, u = nevergrad_ngopt8_cube(horse_dividends_on_cube,n_dim=20, n_trials=25000)
    o = cube_to_ability(u)
    print(' ')
    print('Horse ability, and best solution found ')
    print(list(zip(o, ABILITIES)))


if False:
    from deap.benchmarks import h1
    from humpday.visualization.plotting import mesh2d_
    import numpy as np
    import math

    # Work in progress


    def rotate(origin, point, angle):
        """
        Rotate a point counterclockwise by a given angle around a given origin.

        The angle should be given in radians.
        """
        ox, oy = origin
        px, py = point

        qx = ox + math.cos(angle) * (px - ox) - math.sin(angle) * (py - oy)
        qy = oy + math.sin(angle) * (px - ox) + math.cos(angle) * (py - oy)
        return qx, qy


    def coords(zs,dim=2):
        """ Unflatten into list of tuples """
        return [ np.array(zs[i:i+2]) for i in range(0,len(zs),dim) ]

    def hump(xs,j):
        """ Loosely based on h1 function """
        origin = 0,0
        xs_rotated = rotate(origin=origin,point=xs,angle=j*math.pi/2)
        xs_dilated = [ (x*100) for i,x in enumerate(xs_rotated)]
        return -h1(individual=xs_dilated)[0]+0.0005*np.linalg.norm(xs_dilated)


    def dilated_hump(xs,j):
        """ Loosely based on h1 function """
        origin = 0,0
        xs_rotated = rotate(origin=origin,point=xs,angle=j*math.pi/2)
        xs_dilated = [ (x*100)*(i+1) for i,x in enumerate(xs_rotated)]
        return -h1(individual=xs_dilated)[0]+0.0005*np.linalg.norm(xs_dilated)

    def hump_selector(xs,zs,ws,offset):
        """ Finds nearest center and returns hump function
           xs position in plane
           zs listing of centers of humps
           ws weights of humps
           offset
        """
        centers = coords( zs, 2)
        ys = [ xs-c+offset for c in centers ]
        distances = [ (np.linalg.norm(ys_),i) for i, ys_ in enumerate(ys) ]
        sort_dist = sorted(distances,reverse=False)
        i_close = sort_dist[0][1]
        y = ys[i_close]
        w = ws[i_close]
        return w*hump(y,j=i_close)


if __name__=='__main__':
    if False:
        ws = np.array([1.0,2.0,1.3])
        zs = np.array([0.25, 0.25, -0.25, -0.25, 0.15, -0.11])
        offset = (0,0)
        mesh2d_(hump_selector,zs,ws,offset)


from humpday.objectives.classic import CLASSIC_OBJECTIVES

try:
    from skopt import gp_minimize
    using_gp = True
except ImportError:
    using_gp = False

if using_gp:
    global feval_count
    from copy import deepcopy
    import traceback

    GP_INTERPRETATIONS = dict([(ipg, {'initial_point_generator':ipg}) for ipg in ["random", "sobol", "hammersly", "lhs"]])
    GP_INTERPRETATIONS.update({  'lcbexplore':{"acq_func": "LCB", "kappa":3.0},
                                 'lcb':{"acq_func":"LCB","kappa":1.96},
                                 'lcbexploit':{"acq_func":"LCB","kappa":1.0},
                                 'pi':{"acq_func":"pi"},
                                 'xi02':{"xi":0.02},
                                 'default':{},
                                 'sampling':{'acq_optimizer':'sampling'},
                                 'lbfgs':{'acq_optimizer':'lbfgs'},
                                 'noisy':{'noise':'gaussian'},
                                 'calm':{'noise':1e-8},
                                 '':{}
                               })
    # This could be improved I'm sure
    # https://github.com/scikit-optimize/scikit-optimize/discussions/1002


    def skopt_gp_method_to_kwargs(method:str)->dict:
        """ Create param list for skopt.gp_minimize from name of strategy
             e.g.  lcbexplore_sampling_sobol -> {'acq_func':'LCB',... etc}
        """
        cues = method.split('_')
        params = dict()
        for cue in cues:
            if cue:
                if GP_INTERPRETATIONS.get(cue):
                    params.update(deepcopy(GP_INTERPRETATIONS[cue]))
                else:
                    msg = 'skopt_gp cannot interpret the cue ' + cue
                    raise ValueError(msg)
        return params


    def skopt_gp_cube_factory(objective, n_trials, n_dim, with_count=False, method=''):
        bounds = [(0.,1.,'uniform') ]*n_dim  # <-- Beware: rookie error to accidentally use corners only
                                             # when the dimension isn't explicitly stated as continuous

        global feval_count
        feval_count = 0

        def _objective(x):
            global feval_count
            feval_count +=1
            return objective(list(x))

        gp_params = skopt_gp_method_to_kwargs(method=method)
        result = gp_minimize(func=_objective,  dimensions=bounds, n_calls=n_trials, n_jobs=1, **gp_params)
        best_x = list(result.x)
        best_val = result.fun
        return (best_val, best_x,  feval_count) if with_count else (best_val, best_x)


    # We could now create the following functions programmatically, but it is somewhat easier to debug this way

    def skopt_gp_default_cube(objective, n_trials, n_dim, with_count=False):
        return skopt_gp_cube_factory(objective=objective,n_trials=n_trials,n_dim=n_dim, with_count=with_count,
                                     method='')


    def skopt_gp_sobol_cube(objective, n_trials, n_dim, with_count=False):
            return skopt_gp_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                         method='sobol')


    def skopt_gp_hammersley_cube(objective, n_trials, n_dim, with_count=False):
        return skopt_gp_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                     method='hammersly')


    def skopt_gp_lcb_cube(objective, n_trials, n_dim, with_count=False):
        return skopt_gp_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                     method='lcb')


    def skopt_gp_lcbexplore_cube(objective, n_trials, n_dim, with_count=False):
        return skopt_gp_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                     method='lcbexplore')


    def skopt_gp_lcbexploit_cube(objective, n_trials, n_dim, with_count=False):
        return skopt_gp_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                     method='lcbexploit')


    def skopt_gp_sampling_cube(objective, n_trials, n_dim, with_count=False):
        return skopt_gp_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                     method='sampling')


    SKOPT_GP_CANDIDATES = [ skopt_gp_default_cube, skopt_gp_sobol_cube, skopt_gp_hammersley_cube,
                            skopt_gp_lcb_cube, skopt_gp_lcbexplore_cube, skopt_gp_lcbexploit_cube,
                            skopt_gp_sampling_cube ]

    SKOPT_GP_OPTIMIZERS = [ skopt_gp_default_cube ]
else:
    SKOPT_GP_OPTIMIZERS = []

if __name__ == '__main__':
    always_working = SKOPT_GP_CANDIDATES[:1]
    broken = set()
    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in SKOPT_GP_CANDIDATES:
            try:
                print(optimizer(objective, n_trials=250, n_dim=6, with_count=True))
            except Exception as e:
                print(e)
                broken.add(optimizer)
                always_working.remove(optimizer)
    print(' ')
    print('Sometimes broken: ')
    print([b.__name__ for b in broken])
    print(' ')
    print('Alway working: ')
    print([b.__name__ for b in always_working])






if False:
    import numpy as np
    from humpday.visualization.plotting import mesh2d
    from humpday.objectives.planar_h1 import hump_selector
    from shgo import shgo
    from sklearn.datasets import make_spd_matrix

    # Just messin' around here.
    # Moves h1-like hump functions around on the plane until
    # finding the minima of them because easier.


    def locations_objective(zs, random_ws, offset):
        """ Return mean minimum found
            zs - center of humps
            ws - weights
            offset - offset to apply to hump function
        """
        scores = list()


        for ws_ in random_ws:
            def hs(xs):
                return hump_selector(xs=xs,zs=zs,ws=ws_,offset=offset)
            dim = 2
            bounds = [(-0.5,0.5)]*dim
            res = shgo(hs, bounds, n=10, iters=2)
            score = 100*res.fun + res.nfev
            scores.append(score)
        return np.mean(scores)


    def best_locations(ws,offset, random_ws):
        """
            For given weights on humps, move them around until it helps the search

            z0   Initial locations of humps in flat format [ x1, y1, x2, y2, ... ]
            ws   Weights assigned to each hump
        """

        # Simulate random hump multipliers


        def mb(zs):
            return locations_objective(zs=np.array(zs), random_ws=random_ws, offset=offset)

        dim = int(len(ws))
        bounds = [(-0.5,0.5)]*dim*2
        res = shgo(mb, bounds, n=100, iters=50)
        print(res.fun)
        return res.x


    if __name__=='__main__':

        ws = np.array([0.5,0.5,0.5])

        dim = len(ws)
        zs = 0.333*np.random.rand(2*dim)
        offset = 0.1 * np.random.randn(2)
        zdim = len(ws)
        num_ws = 10

        C = make_spd_matrix(zdim)
        while (C[0][1])**2/(C[0][0]*C[1][1])<0.97:
            C = make_spd_matrix(zdim)
        print(C)
        random_ws = np.random.multivariate_normal(ws, 2*C, num_ws)

        avg = locations_objective(zs=zs, random_ws=random_ws, offset=offset)
        print(avg)

        # Optimize positions of humps making it easier
        zBest = best_locations(ws=ws,offset=offset,random_ws=random_ws)
        print(zBest)
        print(ws)
        import random
        f = lambda x1, x2: hump_selector(xs=np.array([x1, x2]), zs=zBest, ws=random.choice(random_ws), offset=offset)
        mesh2d(f)
        pass






from humpday.objectives.classic import CLASSIC_OBJECTIVES
import warnings
from humpday.transforms.zcurves import curl_factory

# On some systems install is unreliable due to ConfigSpace on some systems,
# so this is not included in setup.

try:
    from ultraopt import fmin
    using_ultraopt = True
except ImportError:
    using_ultraopt = False


if using_ultraopt:

    def ultraopt_cube_factory(objective, n_trials, n_dim, with_count, method):
        global feval_count
        feval_count = 0

        HDL = dict([('u' + str(i), {"_type": "uniform", "_value": [0., 1.]}) for i in range(n_dim)])

        def _objective(config: dict) -> float:
            global feval_count
            feval_count += 1
            u = [config['u' + str(i)] for i in range(n_dim)]
            return objective(u)

        result = fmin(eval_func=_objective, config_space=HDL, optimizer=method, n_iterations=n_trials,
                      n_jobs=1, show_progressbar=False, parallel_strategy="Serial" )
        best_x = [result.best_config['u' + str(i)] for i in range(n_dim)]
        best_val = result.best_loss
        return (best_val, best_x, feval_count) if with_count else (best_val, best_x)


    def ultraopt_etpe_cube(objective, n_trials, n_dim, with_count):
        return ultraopt_cube_factory(objective=objective,n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='ETPE')


    def ultraopt_forest_cube(objective, n_trials, n_dim, with_count):
        return ultraopt_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                     method='Forest')

    def ultraopt_gbrt_cube(objective, n_trials, n_dim, with_count):
        return ultraopt_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                     method='GBRT')

    def ultraopt_etpe_curl2_cube(objective, n_trials, n_dim, with_count):
        # Experimental !
        return curl_factory(optimizer=ultraopt_etpe_cube,objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,d=2)


    def ultraopt_etpe_curl3_cube(objective, n_trials, n_dim, with_count):
        # Experimental !
        return curl_factory(optimizer=ultraopt_etpe_cube, objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, d=3)


    def ultraopt_forest_curl2_cube(objective, n_trials, n_dim, with_count):
        # Experimental !
        return curl_factory(optimizer=ultraopt_forest_cube, objective=objective, n_trials=n_trials, n_dim=n_dim,
                            with_count=with_count, d=2)


    def ultraopt_gbrt_curl2_cube(objective, n_trials, n_dim, with_count):
        # Experimental !
        return curl_factory(optimizer=ultraopt_gbrt_cube, objective=objective, n_trials=n_trials, n_dim=n_dim,
                            with_count=with_count, d=2)

    def ultraopt_random_cube(objective, n_trials, n_dim, with_count):
        return ultraopt_cube_factory(objective=objective,n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='Random')


    def ultraopt_random_curl2_cube(objective, n_trials, n_dim, with_count):
        # Experimental !
        return curl_factory(optimizer=ultraopt_random_cube, objective=objective, n_trials=n_trials, n_dim=n_dim,
                            with_count=with_count, d=2)


    def ultraopt_random_curl3_cube(objective, n_trials, n_dim, with_count):
        # Experimental !
        return curl_factory(optimizer=ultraopt_random_cube, objective=objective, n_trials=n_trials, n_dim=n_dim,
                            with_count=with_count, d=3)


    ULTRAOPT_OPTIMIZERS = [ultraopt_random_cube,
                           ultraopt_forest_cube,
                           ultraopt_gbrt_cube,
                           ultraopt_etpe_cube ]  # <-- the only curled up optimizer likely to make real sense, I suspect
else:
    ULTRAOPT_OPTIMIZERS = []


if __name__=='__main__':
    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in ULTRAOPT_OPTIMIZERS:
            print((optimizer(objective, n_trials=250, n_dim=6, with_count=True)))





class NormHelper():

    # Statistics standard library introduced normal distribution but only in versions above 3.8
    # This adds a tiny amount of backward compatibility, but note that scipy is not a formal dependency so some users
    # will need to install that of their own volition. Python caches imports so don't worry too much about this.

    def __init__(self,**kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def normcdf(x):
        g = NormHelper._normcdf_function()
        return g(x)

    @staticmethod
    def norminv(p):
        f = NormHelper._norminv_function()
        return f(p)

    @staticmethod
    def _norminv_function():
        try:
            from statistics import NormalDist
            return NormalDist(mu=0, sigma=1.0).inv_cdf
        except ImportError:
            from scipy.stats import norm
            return norm.ppf

    @staticmethod
    def _normcdf_function():
        try:
            from statistics import NormalDist
            return NormalDist(mu=0, sigma=1.0).cdf
        except ImportError:
            from scipy.stats import norm
            return norm.cdf




from humpday.transforms.zcurvehelper import ZCurveHelper
from typing import List
import numpy as np
import math

# A mostly failed experiment, thus far.

###########################################################
#                                                         #
# Part I: Constructing an embedding   [0,1]^m -> [0,1]^n  #
#                                                         #
###########################################################


def to_space(p: float, bounds=None, dim: int = None):
    """ Interprets p as a point in a rectangle in R^2 or R^3
         :param bounds  [ (low,high), (low,high), (low,high) ] defaults to unit cube
         :param dim     Dimension. Only used if bounds are not supplied.
    """
    if bounds is None:
        assert dim is not None, "If you don't supply bounds, dimension of hypercube is required"
        bounds = [(0, 1) for _ in range(dim)]
    else:
        dim = len(bounds)

    if dim > 1:
        us = reversed(ZCurveHelper().to_cube(zpercentile=p, dim=dim))  # 0 < us[i] < 1
    else:
        us = [p]
    return [u * (b[1] - b[0]) + b[0] for u, b in zip(us, bounds)]


def from_space(ps: [float], bounds=None) -> float:
    """ [ , ]^n -> [0,1] """
    if bounds is None:
        bounds = [(0, 1) for _ in range(len(ps))]
    us = [(pi - b[0]) / (b[1] - b[0]) for pi, b in zip(ps, bounds)]
    for u in us:
        assert 0 <= u <= 1, "bounds are inconsistent with p=" + str(ps)
    if len(us) > 1:
        return ZCurveHelper().from_cube(list(reversed(us)))
    else:
        return us[0]


def chunk_to_end(l: List, n: int) -> List[List]:
    """ Break list in to evenly sized chunks
        :param n: Size of batches
    """
    rl = list(reversed(l))
    chunks = [list(reversed(rl[x:x + n])) for x in range(0, len(rl), n)]
    return list(reversed(chunks))


def flatten(l: List[List]) -> List:
    return [item for sublist in l for item in sublist]


def curl(u:[float], d:int)->[float]:
    """ Cuts down the dimension by about 1/d
          [0,1]^m -> [0,1]^n
          d :  How many dimensions to fold into 1
    """
    assert d in [2, 3]
    u_chunks = chunk_to_end(u, d)
    return [from_space(uc) for uc in u_chunks]


def uncurl(v:[float], d:int, n_dim:int)->[float]:
    """ Inverse of curl, approximately
         n_dim:
    """
    assert d in [2, 3]
    # First infer the dimension to unfold the first entry into
    u_chunks = chunk_to_end([1 for _ in range(n_dim)], d)
    n_last = len(u_chunks[0])
    if n_last<d:
        u0_dim = len(u_chunks[0])
        # Unfold, u[0] is a special case
        u0 = to_space(v[0], dim=u0_dim)
        u_rest = [to_space(vj, dim=d) for vj in v[1:]]
        return u0 + flatten(u_rest)
    else:
        return flatten([to_space(vj, dim=d) for vj in v])


#######################################################################
#                                                                     #
# Part II: Running optimizers in a lower dimension than the problem   #
#                                                                     #
#######################################################################


def verify_embedding(embedding, inverse, n_dim):
    """ Get the smaller dimension, but also run some checks
    :param embedding: [0,1]^n -> [0,1]^m
    :param inverse:   [0,1]^m -> [0,1]^n
    :return:
    """
    for _ in range(500):
        u0 = list(np.random.rand(n_dim))
        v0 = embedding(u0)
        assert all([0 <= v0_j <= 1 for v0_j in v0]), "pi does not go to the cube"
        u0_check = inverse(v0, n_dim)
        assert np.linalg.norm(np.array(u0_check) - np.array(u0)) < 1e-1, 'map_down does not invert map_up'
    return len(v0)


def embedding_optimizer_factory(optimizer, objective, n_trials, n_dim, with_count, embedding, inverse):
    """ Apply dlib in a lower dimension
           objective: [0,1]^n -> float
           map_down: A bijection from [0,1]^n -> [0,1]^m
           map_up:   The inverse of map_down
    """
    m_dim = verify_embedding(embedding=embedding, inverse=inverse, n_dim=n_dim)


    def _objective(v) -> float:
        """  [0,1]^m -> """
        u = inverse(v, n_dim)
        return objective(u)

    best_val, best_v, feval_count = optimizer(objective=_objective, n_trials=n_trials, n_dim=m_dim, with_count=True)
    best_x = inverse(best_v, n_dim)

    return (best_val, best_x, feval_count) if with_count else (best_val, best_x)


def curl_factory(optimizer, objective, n_trials, n_dim, with_count, d):
    assert n_trials>3*(n_dim+1)
    ranking = importance(objective,n_dim)  # Wastes n_dim+1 trials at least, so pretty stupid
                                           # Just a quick experiment
    ordering = grouped_ordering(ranking,d)
    try:
        inv_ordering = [ ordering.index(j) for j in range(n_dim) ]
    except ValueError:
        print('groan')
        pass

    def permute(u):
        return [ u[j] for j in ordering ]

    def permute_inv(u):
        try:
            return [ u[j] for j in inv_ordering ]
        except IndexError:
            print('what the')
            pass

    debug = True # Remove this crud at some point
    if debug:
        u0 = list(np.random.rand(n_dim))
        u0_check = permute_inv(permute(u0))
        assert np.linalg.norm(np.array(u0)-np.array(u0_check))<1e-6

    def curld(u:[float])->[float]:
        u_permute = permute(u)
        return curl(u_permute,d)

    def uncurld(v:[float], n_dim:int)->[float]:
        u = uncurl(v=v,d=d,n_dim=n_dim)
        return permute_inv(u)

    return embedding_optimizer_factory(optimizer=optimizer, objective=objective, n_trials=n_trials-2*n_dim-2, n_dim=n_dim,
                                       with_count=with_count, embedding=curld, inverse=uncurld)


def dlib_curl_cube3(objective, n_trials, n_dim, with_count):
    """ Curled up version of dlib optimizer """
    return curl_factory(optimizer=dlib_cube, objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, d=3)


#######################################################################
#                                                                     #
# Part III: Dreadful hacks for coordinate importance                  #
#                                                                     #
#######################################################################
from scipy.stats import rankdata


def importance(objective, n_dim):
    # Wasteful and crude
    u1 = [0.3*(x-0.5)+0.5 for x in np.random.rand(n_dim) ]
    u2 = [0.3 * (x - 0.5) + 0.5 for x in np.random.rand(n_dim)]
    f1 = objective(u1)
    f2 = objective(u2)
    abs_derivs = list()
    for j in range(n_dim):
       uj1 = u1
       uj1[j] = uj1[j]+0.1
       fj1 = objective(uj1)
       d1 = (fj1-f1)/0.1
       uj2 = u2
       uj2[j] = uj2[j] + 0.01
       fj2 = objective(uj2)
       d2 = (fj2 - f2)/0.01
       d = 0.5*d1 + 0.5*d2
       abs_derivs.append(abs(d))

    ordering = [ j for d,j in sorted(list(zip( abs_derivs, range(n_dim))),reverse=True) ]
    return ordering


def grouped_ordering(ordering,d):
    """ A permutation of coordinates that might work with curl( ,d)
        e.g. if d=2 this alternates between important and unimportant variables
        The most important variables might not be curled at all.
    """
    n_dim = len(ordering)
    n_group = int(math.floor(n_dim/d))
    n_rem = n_dim-n_group*d

    the_head = ordering[:n_rem]
    the_tail = ordering[n_rem:]

    tail_groups = [ the_tail[j:j + n_group] for j in range(0, len(the_tail), n_group)]
    tail_ordering = flatten(list(map(list, zip(*tail_groups))))
    return the_head + tail_ordering








if __name__ == '__main__':
    from humpday.objectives.classic import deap_combo1_on_cube as an_objective
    from humpday.optimizers.dlibcube import dlib_cube

    best_val, best_x, feval_count = dlib_cube(objective=an_objective, n_trials=50, n_dim=14, with_count=True)
    best_val_0, best_x_0, feval_count_0 = dlib_curl_cube3(objective=an_objective, n_trials=150, n_dim=14,
                                                          with_count=True)
    pass
    # The second version works in higher dimensions
    best_val_2, best_x_2, feval_count_2 = dlib_curl_cube3(objective=an_objective, n_trials=150, n_dim=34,
                                                          with_count=True)



try:
    from bayes_opt import BayesianOptimization
    using_bayesopt = True
except ImportError:
    using_bayesopt = False
    
using_bayesopt = False
print('Bayesian-Optimization turned off pending https://github.com/fmfn/BayesianOptimization/issues/300')

if using_bayesopt:
    from humpday.objectives.classic import CLASSIC_OBJECTIVES
    from humpday.objectives.portfolio import markowitz_realized_on_cube


    BAYESOPT_METHODS = ['ucb', 'ei', 'poi']

    def bayesopt_cube_factory(objective, n_trials, n_dim, with_count,method):

        assert method in BAYESOPT_METHODS

        global feval_count
        feval_count = 0

        pbounds = dict([('u' + str(i), (0., 1.)) for i in range(n_dim)])

        def _neg_objective(**kwargs) -> float:
            global feval_count
            feval_count += 1
            u = [kwargs['u' + str(i)] for i in range(n_dim)]
            return -objective(u)

        optimizer = BayesianOptimization(
            f=_neg_objective,
            pbounds=pbounds,
            verbose=0,
            random_state=1,
        )

        optimizer.maximize(
            init_points=5,
            acq=method,
            n_iter=n_trials - 5,
        )

        best_val = -optimizer.max['target']
        best_x = [optimizer.max['params']['u' + str(i)] for i in range(n_dim)]

        return (best_val, best_x, feval_count) if with_count else (best_val, best_x)


    def bayesopt_ucb_cube(objective, n_trials, n_dim, with_count):
        return bayesopt_cube_factory(objective=objective,n_trials=n_trials,
                                     n_dim=n_dim, with_count=with_count,method='ucb')


    def bayesopt_ei_cube(objective, n_trials, n_dim, with_count):
        return bayesopt_cube_factory(objective=objective,n_trials=n_trials,
                                     n_dim=n_dim, with_count=with_count,method='ei')


    def bayesopt_poi_cube(objective, n_trials, n_dim, with_count):
        return bayesopt_cube_factory(objective=objective,n_trials=n_trials,
                                     n_dim=n_dim, with_count=with_count,method='poi')


    BAYESOPT_OPTIMIZERS = [ bayesopt_ucb_cube, bayesopt_ei_cube, bayesopt_poi_cube ]
else:
    BAYESOPT_OPTIMIZERS = []


if __name__=='__main__':
    assert using_bayesopt
    for objective in [markowitz_realized_on_cube] + CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in BAYESOPT_OPTIMIZERS:
            print((optimizer(objective, n_trials=25, n_dim=3, with_count=True)))



try:
    from pySOT.experimental_design import SymmetricLatinHypercube, LatinHypercube, TwoFactorial
    using_pysot = True
except ImportError:
    using_pysot = False

if using_pysot:

    from poap.controller import BasicWorkerThread, ThreadController
    from pySOT.experimental_design import SymmetricLatinHypercube, LatinHypercube, TwoFactorial
    from pySOT.strategy import SRBFStrategy, EIStrategy, DYCORSStrategy,RandomStrategy, LCBStrategy
    from pySOT.surrogate import CubicKernel, LinearTail, RBFInterpolant, GPRegressor
    from pySOT.optimization_problems.optimization_problem import OptimizationProblem
    import numpy as np
    import logging

    logging.getLogger('pySOT').setLevel(logging.ERROR)


    class GenericProblem(OptimizationProblem):

        def __init__(self, dim, objective):
            self.dim = dim
            self.min = 0
            self.minimum = np.zeros(dim)
            self.lb = 0 * np.ones(dim)
            self.ub = 1 * np.ones(dim)
            self.int_var = np.array([])
            self.cont_var = np.arange(0, dim)
            self.objective = objective
            self.info = str(dim) + "-dimensional objective function " + objective.__name__
            self.feval_count = 0

        def eval(self, x):
            """Evaluate the objective x

            :param x: Data point
            :type x: numpy.array
            :return: Value at x
            :rtype: float
            """
            self.__check_input__(x)
            d = float(self.dim)
            self.feval_count = self.feval_count+1
            return self.objective(list(x))


    def pysot_cube( objective, n_trials, n_dim, with_count=False, method=None, design=None):
        """ Minimize
        :param objective:
        :param n_trials:
        :param n_dim:
        :param with_count:
        :return:
        """
        logging.getLogger('pySOT').setLevel(logging.ERROR)

        num_threads = 1
        asynchronous = True

        max_evals = n_trials
        gp = GenericProblem(dim=n_dim, objective=objective)

        if design=='latin':
            exp_design = LatinHypercube(dim=n_dim, num_pts=2 * (n_dim + 1))
        elif design=='symmetric':
            exp_design = SymmetricLatinHypercube(dim=n_dim, num_pts=2 * (n_dim + 1))
        elif design=='factorial':
            exp_design = TwoFactorial(dim=n_dim)
        else:
            raise ValueError('design should be latin, symmetric or factorial')

        # Create a strategy and a controller
        #  SRBFStrategy, EIStrategy, DYCORSStrategy,RandomStrategy, LCBStrategy
        controller = ThreadController()
        if method.lower()=='srbf':
            surrogate = RBFInterpolant(dim=n_dim, lb=np.array([0.0] * n_dim), ub=np.array([1.0] * n_dim), kernel=CubicKernel(),
                                 tail=LinearTail(n_dim))
            controller.strategy = SRBFStrategy(
                max_evals=max_evals, opt_prob=gp, exp_design=exp_design, surrogate=surrogate, asynchronous=asynchronous
            )
        elif method.lower() == 'ei':
            surrogate = GPRegressor(dim=n_dim, lb=np.array([0.0] * n_dim), ub=np.array([1.0] * n_dim) )
            controller.strategy = EIStrategy(
                max_evals=max_evals, opt_prob=gp, exp_design=exp_design, surrogate=surrogate, asynchronous=asynchronous
            )
        elif method.lower() == 'dycors':
            surrogate = RBFInterpolant(dim=n_dim, lb=np.array([0.0] * n_dim), ub=np.array([1.0] * n_dim), kernel=CubicKernel(),
                                 tail=LinearTail(n_dim))
            controller.strategy = DYCORSStrategy(
                max_evals=max_evals, opt_prob=gp, exp_design=exp_design, surrogate=surrogate, asynchronous=asynchronous
            )
        elif method.lower() == 'lcb':
            surrogate = GPRegressor(dim=n_dim, lb=np.array([0.0] * n_dim), ub=np.array([1.0] * n_dim))
            controller.strategy = LCBStrategy(
                max_evals=max_evals, opt_prob=gp, exp_design=exp_design, surrogate=surrogate, asynchronous=asynchronous
            )
        elif method.lower() == 'random':
            controller.strategy = RandomStrategy(
                max_evals=max_evals, opt_prob=gp
            )
        else:
            raise ValueError("Didn't recognize method passed to pysot")


        # Launch the threads and give them access to the objective function
        for _ in range(num_threads):
            worker = BasicWorkerThread(controller, gp.eval)
            controller.launch_worker(worker)

        # Run the optimization strategy
        result = controller.run()
        best_x = result.params[0].tolist()
        return (result.value, best_x, gp.feval_count) if with_count else (result.value, best_x)


    # Index by strategy
    # SRBFStrategy, EIStrategy, DYCORSStrategy,RandomStrategy, LCBStrategy

    def pysot_srbf_cube( objective, n_trials, n_dim, with_count=False):
        return pysot_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='srbf', design='symmetric' )


    def pysot_ei_cube( objective, n_trials, n_dim, with_count=False):
        return pysot_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='ei', design='symmetric' )


    def pysot_dycors_cube( objective, n_trials, n_dim, with_count=False):
        return pysot_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='dycors', design='symmetric' )


    def pysot_lcb_cube( objective, n_trials, n_dim, with_count=False):
        return pysot_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='lcb', design='symmetric' )


    def pysot_random_cube( objective, n_trials, n_dim, with_count=False):
        return pysot_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, method='random', design='symmetric' )


    PYSOT_OPTIMIZERS = [ pysot_ei_cube, pysot_lcb_cube, pysot_random_cube, pysot_srbf_cube, pysot_dycors_cube ]
    PYSOT_TOP_OPTIMIZERS = [ pysot_srbf_cube, pysot_dycors_cube]
else:
    PYSOT_OPTIMIZERS = []
    PYSOT_TOP_OPTIMIZERS = []


if __name__ == '__main__':
    assert using_pysot
    from humpday.objectives.classic import CLASSIC_OBJECTIVES
    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in PYSOT_OPTIMIZERS:
            print( (optimizer.__name__, optimizer(objective, n_trials=250, n_dim=6, with_count=True)))


from funcy import print_durations

try:
    from ax import optimize
    using_ax = True
except ImportError:
    using_ax = False

if using_ax:

    from logging import CRITICAL
    from ax.utils.common.logger import get_logger
    rt = get_logger('ax')
    rt.setLevel(CRITICAL)
    import warnings
    warnings.filterwarnings("ignore")


    def ax_cube(objective, n_trials, n_dim, with_count=False, method=None):
        global feval_count
        feval_count = 0
        rt = get_logger('ax')
        rt.setLevel(CRITICAL)

        def evaluation_func(prms):
            global feval_count
            feval_count += 1
            return objective([prms["u" + str(i)] for i in range(n_dim)])

        parameters = [{
            "name": "u" + str(i),
            "type": "range",
            "bounds": [0.0, 1.0],
        } for i in range(n_dim)]
        best_parameters, best_values, experiment, model = optimize(parameters=parameters,
                                                                   evaluation_function=evaluation_func,
                                                                   minimize=True,
                                                                   total_trials=n_trials)
        best_x = [ best_parameters['u'+str(i)] for i in range(n_dim) ]
        best_val = best_values[0]['objective']
        return (best_val, best_x, feval_count) if with_count else (best_val, best_x)


    def ax_default_cube(objective, n_trials, n_dim, with_count=False):
        return ax_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count)


    AX_OPTIMIZERS = [ ax_default_cube ]
    AX_TOP_OPTIMIZERS = [ax_default_cube]
else:
    AX_OPTIMIZERS = []
    AX_TOP_OPTIMIZERS = []


if __name__=='__main__':
    from humpday.objectives.classic import CLASSIC_OBJECTIVES
    @print_durations()
    def demo():
        for objective in CLASSIC_OBJECTIVES:
            print(' ')
            print(objective.__name__)
            for optimizer in AX_OPTIMIZERS:
                print(optimizer(objective, n_trials=250, n_dim=6, with_count=True))




from humpday.optimizers.alltopoptimizers import TOP_OPTIMIZERS
from humpday.optimizers.alloptimizers import OPTIMIZERS
from humpday.objectives.classic import CLASSIC_OBJECTIVES
from pprint import pprint
from typing import List
import numpy as np
from humpday.comparison.naming import optimizer_name

# See also https://github.com/microprediction/humpday-testing/tree/main/optimizer_elo_ratings/leaderboards


def compare(objectives:List=None, n_dim=5, n_trials=100):
    """  Run all optimizers and return ranked list
    :param objectives:  A list of functions taking a single argument representing [0,1]^n_dim
    :param n_trials:    Maximum number of function evaluations
    :returns sorted list of minima found and actual number of function evaluations
    """
    if objectives is None:
        objectives = CLASSIC_OBJECTIVES[:2]
    return sorted([(optimizer(objective, n_trials=n_trials, n_dim=n_dim, with_count=True),
                          optimizer_name(optimizer),objective.__name__)
                         for optimizer in OPTIMIZERS for objective in objectives])


def points_race(objectives:List=None, optimizers=None, n_dim=5, n_trials=100, n_top=3):
    """  Run all optimizers and return ranked list, printing as we go
    :param objectives:  A list of functions taking a single argument representing [0,1]^n_dim
    :param n_trials:    Maximum number of function evaluations
    :param n_top:       Determines how many get points awarded, for each obj function (.e.g. 3,2,1 points)
    :returns sorted list of minima found and actual number of function evaluations
    """
    if objectives is None:
        objectives = CLASSIC_OBJECTIVES

    if optimizers is None:
        optimizers = TOP_OPTIMIZERS

    from collections import Counter
    overall_points = None
    for objective in objectives:
        print('Optimizing the '+ objective.__name__.replace('_on_cube','')+' function ...')
        obj_results = list()
        for optimizer in optimizers:
            name = optimizer_name(optimizer)
            best_val, _, feval_count = optimizer(objective, n_trials=n_trials, n_dim=n_dim, with_count=True)
            obj_results.append( (best_val+0.00001*np.random.randn(), name) ) # Tie breaker as I am lazy
        points = Counter( dict([ (name, n_top-j) for j, (val,name) in enumerate( sorted( obj_results )[:n_top] ) ]) )
        if overall_points is None:
            overall_points = points
        else:
            overall_points.update(points)
        print('Best so far ...')
        pprint(overall_points)
    return overall_points






if __name__=='__main__':
    pprint(points_race())







from humpday.objectives.allobjectives import CLASSIC_OBJECTIVES
from humpday.optimizers.scipycube import scipy_powell_cube
import json
from getjson import getjson


# Very approximate benchmark minima (found by scipy.dogleg)

def get_benchmarks():
    return getjson('https://raw.githubusercontent.com/microprediction/humpday/main/humpday/comparison/bencharks.json')


def create_benchmarks():
    benchmarks = dict()
    for objective in CLASSIC_OBJECTIVES:
        benchmarks[objective.__name__] = dict()
        print(objective.__name__)
        for n_dim in [2,3,5,8,13,21,34]:
            benchmarks[objective.__name__][n_dim]=dict()
            for n_trials in [20,30,50,80,130,210,340]:
                best_val,best_x = scipy_powell_cube(objective=objective,n_trials=n_trials,n_dim=n_dim)
                benchmarks[objective.__name__][n_dim][n_trials] = best_val
            with open('bencharks.json','wt') as fp:
                json.dump(benchmarks,fp=fp)

if __name__=='__main__':
    create_benchmarks()




from humpday.objectives.classic import CLASSIC_OBJECTIVES
import logging
import numpy as np
import math
import warnings

try:
    from hebo.design_space.design_space import DesignSpace
    from hebo.optimizers.hebo import HEBO
    using_hebo = True
except ImportError:
    using_hebo = False


if using_hebo:

    logging.getLogger('hebo').setLevel(logging.ERROR)


    def hebo_cube_factory(objective, n_trials, n_dim, with_count,n_suggestions=5):
        global feval_count
        feval_count = 0

        variables = [{'name': 'u' + str(i), 'type': 'num', 'lb': 0., 'ub': 1.} for i in range(n_dim)]
        space = DesignSpace().parse(variables)
        opt = HEBO(space)

        def _objective(params) -> np.ndarray:
            global feval_count
            feval_count += len(params.index)
            return np.array([ objective(ui) for ui in params.values ])

        n_batches = int(math.floor(n_trials/n_suggestions))
        n_remainder = n_trials - n_suggestions*n_batches
        for i in range(n_batches):
            rec = opt.suggest(n_suggestions=n_suggestions) # <-- don't change this
            opt.observe(rec, _objective(rec))
        for i in range(n_remainder):
            rec = opt.suggest(n_suggestions=1)  # <-- don't change this
            opt.observe(rec, _objective(rec))

        best_val = opt.y.min()
        best_ndx = np.argmin([y[0] for y in opt.y])  # I mean seriously, why make the user do this?
        best_x = list(opt.X.values[best_ndx])
        return (best_val, best_x, feval_count) if with_count else (best_val, best_x)


    def hebo_sequential_cube(objective, n_trials, n_dim, with_count):
        return hebo_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, n_suggestions=1)


    def hebo_batch_cube(objective, n_trials, n_dim, with_count):
        return hebo_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count, n_suggestions=10)


    HEBO_OPTIMIZERS = [hebo_sequential_cube, hebo_batch_cube]
else:
    HEBO_OPTIMIZERS = []

if __name__=='__main__':
    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        import time
        for optimizer in HEBO_OPTIMIZERS:
            print(optimizer.__name__+'...')
            st = time.time()
            print((optimizer(objective, n_trials=12, n_dim=4, with_count=True)))
            print('   ... took '+str(time.time()-st)+' seconds.')


from humpday.objectives.classic import CLASSIC_OBJECTIVES, A_CLASSIC_OBJECTIVE
from humpday.objectives.portfolio import PORTFOLIO_OBJECTIVES
from humpday.objectives.horse import HORSE_OBJECTIVES

OBJECTIVES = CLASSIC_OBJECTIVES + PORTFOLIO_OBJECTIVES + HORSE_OBJECTIVES



import numpy as np
import math
from humpday.objectives.deapobjectives import schwefel, schaffer, bohachevsky, griewank, rastrigin, shekel, rosenbrock

# Some test objective functions to help guide optimizer choices
# -------------------------------------------------------------
#
# We'll use DEAP's set of groovy benchmarks, and landscapes, swarmpackagepy also
#
# See pretty pictures at https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks
# Some hardness assessment is at https://github.com/nathanrooy/landscapes#available-functions-from-single_objective but we'll do our own


## Basis of tricky functions

import datetime

DAY = datetime.datetime.today().day
OFFSET = DAY / 50
POWER = 1 + (DAY % 3) / 3.0
SHIFT = DAY / 100


def smoosh(ui):
    """ Distort the interval to avoid obvious minima and avoid memorization """
    ui_rotate = ui + SHIFT % 1.0
    ui_shift = ui_rotate + SHIFT
    xi = ui_shift ** POWER
    low = SHIFT ** POWER
    high = (1 + SHIFT) ** POWER
    yi = (xi - low) / (high - low)
    return yi ** POWER


def schwefel_on_cube(u: [float]) -> float:
    # https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.schwefel
    u_squished = [1000 * (smoosh(ui) - 0.5) for ui in u]
    try:
        return 0.001 * schwefel(u_squished)[0] / 0.71063
    except Exception as e:
        raise Exception(e)


def griewank_on_cube(u: [float]) -> float:
    # https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.griewank
    u_squished = [1200 * (ui ** 1.1 - 0.5) for ui in u]
    return griewank(u_squished)[0] / 0.532075


def rastrigin_on_cube(u: [float]) -> float:
    # https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.rastrigin
    u_squished = [10.24 * (ui ** 1.1 - 0.5) for ui in u]
    return 0.01 * rastrigin(u_squished)[0] / 0.059697


def bohachevsky_on_cube(u: [float]) -> float:
    # https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.bohachevsky
    u_squished = [10 * (ui ** 1.1 - 0.5) for ui in u]
    return 1.0 + bohachevsky(u_squished)[0]


def rosenbrock_on_cube(u: [float]) -> float:
    # https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.rosenbrock
    u_squished = [200 * (ui ** 1.1 - 0.5) for ui in u]
    return 1 + 0.1 * rosenbrock(u_squished)[0] / 0.008949


def shaffer_on_cube(u: [float]) -> float:
    # https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.schaffer
    u_squished = [200 * (ui ** 1.1 - 0.5) for ui in u]
    return 0.01 * schaffer(u_squished)[0] / (0.1042133 * 0.71809)


def shekel_on_cube(u: [float]) -> float:
    # https://deap.readthedocs.io/en/master/api/benchmarks.html#deap.benchmarks.schaffer

    n_dim = len(u)
    NUMMAX = 15
    A = 10 * np.random.rand(NUMMAX, n_dim)
    C = np.random.rand(NUMMAX)
    u_squished = [800 * (smoosh(ui) - 0.5) for ui in u]
    return 1.2298 - shekel(u_squished, A, C)[0]


## Combinations

def deap_combo1_on_cube(u: [float]) -> float:
    return 0.3 * (schwefel_on_cube(u) + griewank_on_cube(u) + shekel_on_cube(u)) / 1.883


def deap_combo2_on_cube(u: [float]) -> float:
    return 0.5 * (shaffer_on_cube(u) + shekel_on_cube(u)) - 0.1075


def deap_combo3_on_cube(u: [float]) -> float:
    return 0.5 * (rosenbrock_on_cube(u) + bohachevsky_on_cube(u) + shekel_on_cube(u)) / 1.88


DEAP_OBJECTIVES = [schwefel_on_cube, rastrigin_on_cube, griewank_on_cube,
                   bohachevsky_on_cube, rosenbrock_on_cube, shaffer_on_cube, shekel_on_cube,
                   deap_combo1_on_cube, deap_combo2_on_cube, deap_combo3_on_cube]


# By hand...

def rosenbrock_modified_on_cube(u: [float]) -> float:
    """ https://en.wikipedia.org/wiki/Rosenbrock_function """
    u_scaled = [4 * ui - 2 for ui in u]
    if len(u) == 1:
        return (0.25 - u_scaled[0]) ** 2
    else:
        return 5 + 0.001 * np.sum(
            [100 * (ui_plus - ui * ui) + (1 - ui) * (1 - ui) for ui, ui_plus in zip(u_scaled[1:], u_scaled)])


# According to http://infinity77.net/global_optimization/test_functions.html#test-functions-index
# there are some really hard ones
# See https://github.com/andyfaff/ampgo/blob/master/%20ampgo%20--username%20andrea.gavana%40gmail.com/go_benchmark.py
# See also https://arxiv.org/pdf/1308.4008v1.pdf

def damavandi_on_cube(u: [float]) -> float:
    """ A trivial multi-dimensional extension of Damavandi's function """
    return 0.01 * damavandi2(u[0], u[1]) - 0.46


def damavandi2(u1, u2) -> float:
    """ Pretty evil function this one """
    # http://infinity77.net/global_optimization/test_functions_nd_D.html#go_benchmark.Damavandi
    x1 = u1 / 14.
    x2 = u2 / 14.
    numerator = math.sin(math.pi * (x1 - 2.0)) * math.sin(math.pi * (x2 - 2.0))
    denumerator = (math.pi ** 2) * (x1 - 2.0) * (x2 - 2.0)
    factor1 = 1.0 - (abs(numerator / denumerator)) ** 5.0
    factor2 = 2 + (x1 - 7.0) ** 2.0 + 2 * (x2 - 7.0) ** 2.0
    return factor1 * factor2


def paviani_on_cube(u: [float]) -> float:
    # http://infinity77.net/global_optimization/test_functions_nd_P.html#go_benchmark.Paviani
    x = np.array([2.001 + 5.996 * smoosh(ui) for ui in u])

    def safe_np_log(x):
        lb = np.array([1e-6] * len(x))
        xup = np.maximum(x, lb)
        return np.log(xup)

    return float(np.sum(safe_np_log(x - 2) ** 2.0 + safe_np_log(10.0 - x) ** 2.0) - np.prod(x) ** 0.2) / 8.6456


# Landscapes

from landscapes.single_objective import styblinski_tang, zakharov, salomon, rotated_hyper_ellipsoid, qing, michalewicz


def styblinski_tang_on_cube(u: [float]) -> float:
    u_scaled = [10 * (smoosh(ui) - 0.5) for ui in u]
    return 3.3499 + 0.01 * styblinski_tang(u_scaled)


def zakharov_on_cube(u: [float]) -> float:
    u_scaled = [15 * smoosh(ui) - 10 for ui in u]
    return 0.01 * zakharov(u_scaled) / 0.3462


def salomon_on_cube(u: [float]) -> float:
    u_scaled = [200 * smoosh(ui) - 100 for ui in u]
    return salomon(u_scaled) / 3.09999


def rotated_hyper_ellipsoid_on_cube(u: [float]) -> float:
    u_scaled = [2 * 65.536 * smoosh(ui) - 65.536 for ui in u]
    return 0.1 * rotated_hyper_ellipsoid(u_scaled)


def qing_on_cube(u: [float]) -> float:
    u_scaled = [1000 * smoosh(ui) - 500 for ui in u]
    return qing(u_scaled) / 0.01805


def michaelewicz_on_cube(u: [float]) -> float:
    u_scaled = [4 * smoosh(ui) - 2 for ui in u]
    return 1.4439 + 0.1 * michalewicz(u_scaled, m=20)


def landscapes_combo1_on_cube(u: [float]) -> float:
    return (qing_on_cube(u) + michaelewicz_on_cube(u)) / (1.5744 * 1.4688)


def landscapes_combo2_on_cube(u: [float]) -> float:
    return (rotated_hyper_ellipsoid_on_cube(u) + salomon_on_cube(u)) / (6.7555 * 0.82)


def landscapes_combo3_on_cube(u: [float]) -> float:
    return (2 + zakharov_on_cube(u) + styblinski_tang_on_cube(u)) / 4.4329


LANDSCAPES_OBJECTIVES = [styblinski_tang_on_cube, zakharov_on_cube, salomon_on_cube, rotated_hyper_ellipsoid_on_cube,
                         qing_on_cube, michaelewicz_on_cube, landscapes_combo1_on_cube, landscapes_combo2_on_cube,
                         landscapes_combo3_on_cube]


# Some copied from peabox
# https://github.com/stromatolith/peabox/blob/master/peabox/peabox_testfuncs.py
# as that isn't deployed to PyPI as far as I can determine


def ackley_on_cube(u: [float]) -> float:
    # allow parameter range -32.768<=x(i)<=32.768, global minimum at x=(0,0,...,0)
    rescaled_u = [2 * 32.768 * smoosh(ui) - 32.768 for ui in u]
    x = np.asfarray(rescaled_u)
    ndim = len(x)
    a = 20.;
    b = 0.2;
    c = 2. * math.pi
    return (-a * np.exp(-b * np.sqrt(1. / ndim * np.sum(x ** 2))) - np.exp(
        1. / ndim * np.sum(np.cos(c * x))) + a + np.exp(1.)) / 20.0


# Adapted from https://github.com/SISDevelop/SwarmPackagePy/blob/master/SwarmPackagePy/testFunctions.py

def cross_on_cube(u):
    x = [5 * smoosh(ui) - 2.5 for ui in u]
    return round(-0.0001 * (abs(math.sin(x[0]) * math.sin(x[1]) * math.exp(abs(100 -
                                                                               math.sqrt(sum([i ** 2 for i in
                                                                                              x])) / math.pi))) + 1) ** 0.1,
                 7)


def powers_on_cube(u):
    x = [5 * smoosh(ui) - 2.5 for ui in u]
    return sum([abs(x[i]) ** (i + 2) for i in range(len(x))])


def booth_on_cube(u):
    x = [5 * smoosh(ui) - 2.5 for ui in u]
    return sum([abs(x[i]) ** (i + 2) for i in range(len(x))])


def matyas_on_cube(u):
    x = [3 * smoosh(ui) - 1.5 for ui in u]

    def sphere_function(x):
        return sum([i ** 2 for i in x])

    return 0.26 * sphere_function(x) - 0.48 * x[0] * x[1]


def drop_wave_on_cube(u):
    x = [3 * smoosh(ui) - 1.5 for ui in u]

    def sphere_function(x):
        return sum([i ** 2 for i in x])

    return -(1 + math.cos(12 * math.sqrt(sphere_function(x)))) / (0.5 * sphere_function(x) + 2)


SWARM_OBJECTIVES = [cross_on_cube, powers_on_cube,
                    booth_on_cube, matyas_on_cube, drop_wave_on_cube]

A_CLASSIC_OBJECTIVE = rastrigin_on_cube  # Just pick one for testing

MISC_OBJECTIVES = [paviani_on_cube, damavandi_on_cube, rosenbrock_modified_on_cube, ackley_on_cube]

CLASSIC_OBJECTIVES = DEAP_OBJECTIVES + LANDSCAPES_OBJECTIVES + MISC_OBJECTIVES + SWARM_OBJECTIVES

if __name__ == "__main__":
    for objective in CLASSIC_OBJECTIVES:
        objective(u=[0.0, 0.5, 1.0])
        objective(u=[0.0, 0.5, 0.0, 0.0, 1.0])
    print(len(CLASSIC_OBJECTIVES))



from getjson import getjson


ELO_URL='https://raw.githubusercontent.com/microprediction/optimizer-elo-ratings/main/results/CATEGORY_dDD_nNN.json'


def elo_leaderboard_url(category='classic',n_dim=5, n_trials=130):
    return ELO_URL.replace('CATEGORY',category.lower()).replace('DD',str(n_dim).zfill(2)).replace('NN',str(n_trials))


def get_elo_leaderboard(category='classic',n_dim=5, n_trials=130):
    return getjson(elo_leaderboard_url(category=category,n_dim=n_dim,n_trials=n_trials))


def get_elo_rating(strategy='scipy_powell', category='classic',n_dim=5, n_trials=130):
    s = strategy+'_cube' if '_cube' not in strategy else strategy
    lb = get_elo_leaderboard(category=category,n_dim=n_dim,n_trials=n_trials)
    try:
        ndx = lb['name'].index(s)
        return lb['rating'][ndx]
    except ValueError:
        return None








if __name__=='__main__':
    print(get_elo_rating('pysot_ei_cube'))


import math
from humpday.comparison.suggestions import recommend
from humpday.optimizers.alloptimizers import optimizer_from_name

# The equivalent of the "I'm feeling lucky" button on Google search. 


def minimize(objective,n_dim:int,n_trials:int, category='classic', with_count=False):
    """ Minimize an objective, by first choosing a good optimizer then using it.
    :param objective:
    :param n_dim:
    :param n_trials:
    :param category:
    :param with_count:
    :return:
    """
    recommendations = recommend(objective=objective, n_dim=n_dim, n_trials=n_trials, category=category )
    opt = recommendations[0][2]
    optim = optimizer_from_name(opt)
    print('Using '+opt)
    return optim(objective, n_dim=n_dim, n_trials=n_trials, with_count=with_count)


if __name__=='__main__':
    import time
    import math
    from pprint import pprint

    def my_objective(u):
        time.sleep(0.1)
        return u[0]*math.sin(u[1])

    pprint(minimize(my_objective, n_dim=3, n_trials=130, with_count=True))



try:
    from pymoo.optimize import minimize
    from pymoo.model.problem import Problem
    using_pymoo = True
except ImportError:
    using_pymoo = False

if using_pymoo:

    from pymoo.optimize import minimize
    from pymoo.model.problem import Problem
    import numpy as np
    from pymoo.factory import get_algorithm, get_termination, get_reference_directions
    import logging
    logging.getLogger('pymoo').setLevel(logging.ERROR)
    from humpday.transforms.zcurves import curl_factory

    REF_DIRS = get_reference_directions("das-dennis", 1, n_partitions=6)


    def pymoo_brkga_cube(objective, n_trials, n_dim, with_count=False):
        return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="brkga", n_dim=n_dim, with_count=with_count)


    def pymoo_nelder_cube(objective, n_trials, n_dim, with_count=False):
        return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="nelder-mead", n_dim=n_dim, with_count=with_count)


    def pymoo_cmaes_cube(objective, n_trials, n_dim, with_count=False):
        return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="cmaes", n_dim=n_dim, with_count=with_count)


    def pymoo_nsga2_cube(objective, n_trials, n_dim, with_count=False):
        return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="nsga2", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


    def pymoo_rnsga2_cube(objective, n_trials, n_dim, with_count=False):
        return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="rnsga2", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


    def pymoo_rnsga3_cube(objective, n_trials, n_dim, with_count=False):
        return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="rnsga3", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


    def pymoo_unsga3_cube(objective, n_trials, n_dim, with_count=False):
        return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="unsga3", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


    def pymoo_moead_cube(objective, n_trials, n_dim, with_count=False):
        return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="moead", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


    def pymoo_pattern_cube(objective, n_trials, n_dim, with_count=False):
        return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="pattern-search", n_dim=n_dim, with_count=with_count)


    def pymoo_ctaea_cube(objective, n_trials, n_dim, with_count=False):
        return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="ctaea", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


    def pymoo_nsga3_cube(objective, n_trials, n_dim, with_count=False):
        return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="nsga3", ref_dirs=REF_DIRS, n_dim=n_dim, with_count=with_count)


    def pymoo_de_cube(objective, n_trials, n_dim, with_count=False):
        return pymoo_cube(objective=objective,    n_trials=n_trials, method_name="nsga3", n_dim=n_dim, with_count=with_count)


    def pymoo_cube(objective, n_trials, method_name, n_dim, with_count, ref_dirs=None):

        class ObjectiveProblem(Problem):

            def __init__(self):
                super().__init__(n_var=n_dim, n_obj=1, n_constr=0, xl=0.0, xu=1.0)
                self.feval_count = 0

            def _evaluate(self, x, out, *args, **kwargs):
                """ vectorized  """
                self.feval_count = self.feval_count + len(x)
                out["F"] = np.array([objective(u) for u in x])

        try:
            algorithm = get_algorithm(method_name, ref_dirs=ref_dirs)
        except ValueError:
            algorithm = get_algorithm(method_name)
        termination = get_termination("n_eval", n_trials)
        problem = ObjectiveProblem()

        result = minimize(problem=problem,
                          algorithm=algorithm,
                          termination=termination,
                          seed=None,
                          verbose=False,
                          display=None,
                          callback=None,
                          return_least_infeasible=False,
                          save_history=False
                          )
        best_val = result.F[0]
        best_x = result.X.tolist()
        return (best_val, best_x, problem.feval_count) if with_count else (best_val, best_x)


    def pymoo_unsga3_curl2_cube(objective, n_trials, n_dim, with_count=False):
        # Mostly failed experiment
        return curl_factory(optimizer=pymoo_unsga3_cube, objective=objective,
                            n_trials=n_trials, n_dim=n_dim, with_count=with_count, d=2)



    PYMOO_CANDIDATES = [ pymoo_de_cube, pymoo_nsga2_cube, pymoo_rnsga2_cube,
                         pymoo_nelder_cube, pymoo_ctaea_cube,
                         pymoo_nsga3_cube, pymoo_rnsga3_cube,
                         pymoo_pattern_cube, pymoo_brkga_cube, pymoo_nsga2_cube,
                         pymoo_unsga3_cube, pymoo_unsga3_curl2_cube]

    BAD = [pymoo_brkga_cube, pymoo_nsga2_cube ] # TODO:  investigate


    PYMOO_OPTIMIZERS = [ pymoo_nelder_cube, pymoo_nsga3_cube,
                       pymoo_unsga3_cube, pymoo_pattern_cube,  pymoo_nsga2_cube,
                        pymoo_unsga3_cube ]
    PYMOO_TOP_OPTIMIZERS = [ pymoo_pattern_cube, pymoo_nelder_cube]
else:
    PYMOO_OPTIMIZERS = []
    PYMOO_TOP_OPTIMIZERS = []

# TODO: See why ['pymoo_ctaea_cube', 'pymoo_de_cube', 'pymoo_rnsga2_cube', 'pymoo_rnsga3_cube'] are broken sometimes

if __name__ == '__main__':
    assert using_pymoo
    from humpday.objectives.classic import CLASSIC_OBJECTIVES
    always_working = PYMOO_CANDIDATES
    broken = set()
    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in PYMOO_CANDIDATES:
            try:
                print(optimizer(objective, n_trials=250, n_dim=15, with_count=True))
            except Exception as e:
                print(e)
                broken.add(optimizer)
                always_working.remove(optimizer)
    print(' ')
    print('Sometimes broken: ')
    print([b.__name__ for b in broken])
    print(' ')
    print('Alway working: ')
    print([b.__name__ for b in always_working])




import math
from typing import Union, Tuple, List
BOUNDS_TYPE = List[Union[Tuple, List]]  # scipy.optimize style bounds [ (low,high), (low, high),... ]


# Misc mappings that might be useful


def positive_log_scale(u, low, high):
    """ Map u in (0,1) to (low,high) """
    assert 0 <= u <= 1
    assert 0 < low < high
    log_low = math.log(low)
    log_high = math.log(high)
    x = log_low + u * (log_high - log_low)
    return math.exp(x)


def to_log_space_1d(u, low, high):
    """ Approximately logarithmic map, but allows for ranges spanning zero
         returns:  float between low and high
    """
    assert 0 <= u <= 1
    assert low<high

    if 1e-8 < low < high:
        return positive_log_scale(u=u, low=low, high=high)
    elif low < -1e-8 < high < 1e-8:
        return -positive_log_scale(1 - u, low=-high, high=-low)
    elif -1e-8 < low < 1e-8 < high:
        return positive_log_scale(u=u, low=1e-8, high=high)
    elif low < -1e-8 < high < 1e-8:
        return -positive_log_scale(1 - u, low=1e-8, high=-low)
    else:
        scale = abs(high - low) / 100
        if u < 0.475:
            u1 = 1 - u / 0.475
            return -positive_log_scale(u=u1, low=scale, high=-low)
        elif 0.475 < u < 0.525:
            u2 = 20 * (u - 0.475)
            return -scale + 2 * u2 * scale
        else:
            u3 = (u - 0.525) / 0.525
            return positive_log_scale(u3, low=scale, high=high)




try:
    from platypus import NSGAII, Problem, Real, EvolutionaryStrategy, GeneticAlgorithm, \
    NSGAIII, CMAES, GDE3, IBEA, MOEAD, OMOPSO, SMPSO, SPEA2, EpsMOEA, normal_boundary_weights
    using_platypus = True
except ImportError:
    using_platypus = False


if using_platypus:
    from platypus import NSGAII, Problem, Real, EvolutionaryStrategy, GeneticAlgorithm, \
        NSGAIII, CMAES, GDE3, IBEA, MOEAD, OMOPSO, SMPSO, SPEA2, EpsMOEA, normal_boundary_weights
    from platypus.core import FixedLengthArray
    import logging

    logging.getLogger('Platypus').setLevel(logging.ERROR)

    # Platypus is designed for multi-objective optimization, so may not do so well on single-objective problems

    PLATYPUS_ALGORITHMS = {'evolutionary': EvolutionaryStrategy,
                           'genetic': GeneticAlgorithm,
                           'nsgaii': NSGAII,
                           'nsgaiii': (NSGAIII, {"divisions_outer": 12}),
                           'cmaes': (CMAES, {"epsilons": [0.05]}),
                           'gde3': GDE3,
                           'ibea': IBEA,
                           'moead': (MOEAD, {"weight_generator": normal_boundary_weights, "divisions_outer": 12}),
                           'omopso': (OMOPSO, {"epsilons": [0.05]}),
                           'smpso': SMPSO,
                           'spea2': SPEA2,
                           'epsmoea': (EpsMOEA, {"epsilons": [0.05]})
                           }


    def platypus_cube(objective, n_trials, n_dim, with_count=False, method=None):
        global feval_count
        feval_count = 0

        def _objective(vars):
            global feval_count
            feval_count += 1
            return float(objective(list(vars)))  # Avoid np.array as Platypus may puke

        problem = Problem(n_dim, 1, 0)
        problem.types[:] = [Real(0.0, 1.0)] * n_dim
        problem.constraints[:] = "<=0"
        problem.function = _objective

        strategy_and_args = PLATYPUS_ALGORITHMS[method]
        if isinstance(strategy_and_args, tuple):
            strategy = strategy_and_args[0]
            strategy_args = strategy_and_args[1]
            algorithm = strategy(problem, **strategy_args)
        else:
            strategy = strategy_and_args
            algorithm = strategy(problem)

        algorithm.run(n_trials)
        feasible_solution_obj = sorted([(s.objectives[0], s.variables) for s in algorithm.result if s.feasible],
                                       reverse=False)
        best_obj, best_x = feasible_solution_obj[0]
        if isinstance(best_x, FixedLengthArray):
            best_x = best_x._data  # CMA-ES returns it this way for some reason
        return (best_obj, best_x, feval_count) if with_count else (best_obj, best_x)


    def platypus_evolutionary_cube(objective, n_trials, n_dim, with_count=False):
        return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                             method='genetic')


    def platypus_genetic_cube(objective, n_trials, n_dim, with_count=False):
        return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                             method='genetic')


    def platypus_nsgaii_cube(objective, n_trials, n_dim, with_count=False):
        return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                             method='nsgaii')


    def platypus_cmaes_cube(objective, n_trials, n_dim, with_count=False):
        return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                             method='cmaes')


    def platypus_gde3_cube(objective, n_trials, n_dim, with_count=False):
        return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                             method='gde3')


    def platypus_ibea_cube(objective, n_trials, n_dim, with_count=False):
        return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                             method='ibea')


    def platypus_moead_cube(objective, n_trials, n_dim, with_count=False):
        return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                             method='moead')


    def platypus_omopso_cube(objective, n_trials, n_dim, with_count=False):
        return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                             method='omopso')


    def platypus_smpso_cube(objective, n_trials, n_dim, with_count=False):
        return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                             method='smpso')


    def platypus_spea2_cube(objective, n_trials, n_dim, with_count=False):
        return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                             method='spea2')


    def platypus_epsmoea_cube(objective, n_trials, n_dim, with_count=False):
        return platypus_cube(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                             method='epsmoea')


    BAD  = [ platypus_nsgaii_cube ]

    PLATYPUS_OPTIMIZERS = [platypus_genetic_cube, platypus_evolutionary_cube,
                           platypus_nsgaii_cube, platypus_cmaes_cube, platypus_gde3_cube,
                           platypus_ibea_cube, platypus_moead_cube, platypus_omopso_cube,
                           platypus_smpso_cube, platypus_spea2_cube, platypus_epsmoea_cube]
else:
    PLATYPUS_OPTIMIZERS =[]

if __name__ == '__main__':
    assert using_platypus
    from humpday.objectives.classic import CLASSIC_OBJECTIVES

    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in PLATYPUS_OPTIMIZERS:
            print((optimizer.__name__, optimizer(objective, n_trials=250, n_dim=6, with_count=True)))



# So horrendous this needs a rewrite

from humpday.optimizers.nevergradcube import NEVERGRAD_TOP_OPTIMIZERS
from humpday.optimizers.optunacube import OPTUNA_TOP_OPTIMIZERS
from humpday.optimizers.nloptcube import NLOPT_TOP_OPTIMIZERS
from humpday.optimizers.dlibcube import DLIB_TOP_OPTIMIZERS
from humpday.optimizers.pysotcube import PYSOT_TOP_OPTIMIZERS
from humpday.optimizers.pymoocube import PYMOO_TOP_OPTIMIZERS
from humpday.optimizers.freelunchcube import FREELUNCH_TOP_OPTIMIZERS
from humpday.optimizers.bobyqacube import BOBYQA_TOP_OPTIMIZERS
from humpday.optimizers.axcube import AX_TOP_OPTIMIZERS
from humpday.optimizers.scipycube import SCIPY_TOP_OPTIMIZERS

TOP_OPTIMIZERS = NEVERGRAD_TOP_OPTIMIZERS + OPTUNA_TOP_OPTIMIZERS + NLOPT_TOP_OPTIMIZERS\
                 + DLIB_TOP_OPTIMIZERS + PYSOT_TOP_OPTIMIZERS + PYMOO_TOP_OPTIMIZERS\
                 + FREELUNCH_TOP_OPTIMIZERS + BOBYQA_TOP_OPTIMIZERS + AX_TOP_OPTIMIZERS+ SCIPY_TOP_OPTIMIZERS









import pymorton
import math
from humpday.transforms.normhelper import NormHelper
from typing import List


class ZCurveHelper():
    """ Conventions for projections R^2->R and R^3->R """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


    @staticmethod
    def to_zscores(prctls):
        norminv = NormHelper._norminv_function()
        return [norminv(p) for p in prctls]

    @staticmethod
    def morton_scale(dim):
        return 2**10

    @staticmethod
    def morton_large(dim):
        SCALE = ZCurveHelper.morton_scale(dim=dim)
        return pymorton.interleave( *[ SCALE-1 for _ in range(dim) ] )

    def to_zcurve(self, prctls: List[float] ):
        """ A mapping from I^n -> R based on the Morton z-curve """
        SAFE = False
        dim = len(prctls)
        if dim==1:
            return self.to_zscores(prctls)[0]
        else:
            zpercentile = self.from_cube(prctls=prctls)
            return NormHelper.norminv(zpercentile)

    def from_cube(self, prctls: List[float]) -> float:
        """
            Inverse of space filling curve  I^n -> I
        """
        dim = len(prctls)
        SAFE = False # relic
        SCALE = self.morton_scale(dim)
        int_prctls = [int(math.floor(p * SCALE)) for p in prctls]
        m1 = pymorton.interleave(*int_prctls)
        if SAFE:
            int_prctls_back = pymorton.deinterleave2(m1) if dim == 2 else pymorton.deinterleave3(m1)
            assert all(i1 == i2 for i1, i2 in zip(int_prctls, int_prctls_back))
        m2 = pymorton.interleave(*[SCALE - 1 for _ in range(dim)])
        zpercentile = m1 / m2
        return zpercentile

    def to_cube(self, zpercentile:float, dim):
        """
            Space filling curve I -> I^n
        """
        SCALE = self.morton_scale(dim)
        zmorton = int(self.morton_large(dim) * zpercentile + 0.5)
        if dim == 2:
            values = pymorton.deinterleave2(zmorton)
        elif dim == 3:
            values = pymorton.deinterleave3(zmorton)
        else:
            raise NotImplementedError('Only 2d or 3d')
        prtcls = [v / SCALE for v in values]
        return prtcls

    def from_zcurve(self, zvalue, dim):
        zpercentile = NormHelper.normcdf(zvalue)
        return self.to_cube(zpercentile=zpercentile, dim=dim)




# DEAP package uses 2to3 which isn't supported
# This is causing problems, so the deap benchmarks are copied here for the time being
# Many thanks to the DEAP team for compiling them.

import random
from math import sin, cos, pi, exp, e, sqrt
from operator import mul
from functools import reduce


def rand(individual):
    return random.random(),


def plane(individual):
    return individual[0],


def sphere(individual):
    return sum(gene * gene for gene in individual),


def cigar(individual):
    return individual[0] ** 2 + 1e6 * sum(gene * gene for gene in individual),


def rosenbrock(individual):
    return sum(100 * (x * x - y) ** 2 + (1. - x) ** 2
               for x, y in zip(individual[:-1], individual[1:])),


def h1(individual):
    num = (sin(individual[0] - individual[1] / 8)) ** 2 + (sin(individual[1] + individual[0] / 8)) ** 2
    denum = ((individual[0] - 8.6998) ** 2 + (individual[1] - 6.7665) ** 2) ** 0.5 + 1
    return num / denum,


def ackley(individual):
    N = len(individual)
    return 20 - 20 * exp(-0.2 * sqrt(1.0 / N * sum(x ** 2 for x in individual))) \
           + e - exp(1.0 / N * sum(cos(2 * pi * x) for x in individual)),


def bohachevsky(individual):
    return sum(x ** 2 + 2 * x1 ** 2 - 0.3 * cos(3 * pi * x) - 0.4 * cos(4 * pi * x1) + 0.7
               for x, x1 in zip(individual[:-1], individual[1:])),


def griewank(individual):
    return 1.0 / 4000.0 * sum(x ** 2 for x in individual) - \
           reduce(mul, (cos(x / sqrt(i + 1.0)) for i, x in enumerate(individual)), 1) + 1,


def rastrigin(individual):
    return 10 * len(individual) + sum(gene * gene - 10 *
                                      cos(2 * pi * gene) for gene in individual),


def rastrigin_scaled(individual):
    N = len(individual)
    return 10 * N + sum((10 ** (i / (N - 1)) * x) ** 2 -
                        10 * cos(2 * pi * 10 ** (i / (N - 1)) * x) for i, x in enumerate(individual)),


def rastrigin_skew(individual):
    N = len(individual)
    return 10 * N + sum((10 * x if x > 0 else x) ** 2
                        - 10 * cos(2 * pi * (10 * x if x > 0 else x)) for x in individual),


def schaffer(individual):
    return sum((x ** 2 + x1 ** 2) ** 0.25 * ((sin(50 * (x ** 2 + x1 ** 2) ** 0.1)) ** 2 + 1.0)
               for x, x1 in zip(individual[:-1], individual[1:])),


def schwefel(individual):
    N = len(individual)
    return 418.9828872724339 * N - sum(x * sin(sqrt(abs(x))) for x in individual),


def himmelblau(individual):
    return (individual[0] * individual[0] + individual[1] - 11) ** 2 + \
           (individual[0] + individual[1] * individual[1] - 7) ** 2,


def shekel(individual, a, c):
    return sum((1. / (c[i] + sum((individual[j] - aij) ** 2 for j, aij in enumerate(a[i])))) for i in range(len(c))),




# Create a plot of strength versus speed
from humpday.comparison.timingretrieval import get_timing
from humpday.comparison.eloretrieval import get_elo_leaderboard
from typing import List, Tuple
from humpday.optimizers.scipycube import scipy_powell_cube as DEFAULT_OPTIMIZER
import time
from pprint import pprint
import math

FIBONACCI = [ 2, 3, 5, 8, 13, 21, 34, 55, 89, 144, 233, 377, 610, 987, 1597]
FIBONACCI_X10 = [ k*10 for k in FIBONACCI ]


def closest(lst, k):
    return lst[min(range(len(lst)), key=lambda i: abs(lst[i] - k))]


def f_dim(n_dim):
    """ Nearest Fibonacci """
    return closest(FIBONACCI,n_dim)


def f_trials(n_trials):
    """ Nearest 10x multiple of Fibonacci """
    return closest(FIBONACCI_X10, n_trials)


def suggest(n_dim:int, n_trials:int, n_seconds:float, category='classic')->List[Tuple]:
    """
         n_seconds:  Maximum allowed time for entire minimization routine
    """

    n_dim = f_dim(n_dim)
    n_trials = f_trials(n_trials)

    cpu = get_timing()
    shortlist = list()
    for opt, opt_stats in cpu.items():
        for dim, dim_stats in opt_stats.items():
            if int(dim)==n_dim:
                if dim_stats.get(str(n_trials)) is not None and dim_stats[str(n_trials)]<=n_seconds:
                    shortlist.append((opt,dim_stats[str(n_trials)]))

    lb = get_elo_leaderboard(category=category, n_dim=n_dim, n_trials=n_trials)
    suggestions = list()
    if lb:
        for opt,t in shortlist:
            try:
                ndx = lb['name'].index(opt)
                elo = lb['rating'][ndx]
                suggestions.append((elo,t,opt))
            except ValueError:
                pass

    if not suggestions:
        return [(None, None, DEFAULT_OPTIMIZER.__name__)]
    else:
        return list(sorted( suggestions, reverse=True))


def recommend(objective, n_dim:int, n_trials:int, category='classic'):
    """
    :param objective:    function taking  list -> float
    :param n_dim:
    :param n_trials:
    :param category:
    :return: List of recommended optimizers
    """
    st = time.time()
    for _ in range(2):
        u = [0.5]*n_dim
        _ = objective(u)
    t = (time.time()-st)/2.0
    n_seconds = 0.1*t*n_trials
    return suggest(n_dim=n_dim, n_trials=n_trials, n_seconds=n_seconds, category=category)


if __name__=='__main__':

    def my_objective(u):
        time.sleep(0.1)
        return u[0]*math.sin(u[1])

    pprint(recommend(my_objective, n_dim=4, n_trials=120)[:10])



import numpy as np
from sklearn.datasets import make_spd_matrix
import time
from scipy.stats import skew
import math

# This problem is randomly generated, but only once per day
# If a solution run straddles GMT midnight, well that's life in the big city.

global sigma_matrix
sigma_matrix = None
global sigma_scenarios
sigma_scenarios = None
global adjusted_scenarios
adjusted_scenarios = None
N_ASSETS = 233
N_SCENARIOS = 15000
global solutions
solutions = None
DT = 1/2000  #
SQDT = math.sqrt(DT)
YEARLY = 3.0


def make_sigma_matrix():
    from datetime import datetime
    day_of_year = datetime.now().timetuple().tm_yday
    return make_spd_matrix(n_dim=N_ASSETS, random_state=day_of_year)


def make_sigma_scenarios():
    """
    :return:  N_SCENARIOS X FIVE_HUNDRED
    """
    global sigma_scenarios
    if sigma_scenarios is None:
        global sigma_matrix
        if sigma_matrix is None:
            sigma_matrix = make_sigma_matrix()
        mu = np.zeros(N_ASSETS)
        sigma_scenarios = SQDT*YEARLY*np.random.multivariate_normal(mu, sigma_matrix, N_SCENARIOS, check_valid='warn', tol=1e-8)
    return sigma_scenarios


def make_adjusted_scenarios():
    """
       :return:
    """
    global adjusted_scenarios
    if adjusted_scenarios is None:
        sigma_scenarios = make_sigma_scenarios()
        mean_ratio = np.mean(np.exp(sigma_scenarios),axis=0)
        mean_expon = np.log(mean_ratio)
        adjusted_scenarios = sigma_scenarios.copy()
        n_dim = sigma_scenarios.shape[1]
        for j in range(n_dim):
            adjusted_scenarios[:,j] = sigma_scenarios[:,j]-mean_expon[j]
        # Blach
        if False: # Later...,
            realized_sigma = np.cov(np.transpose(sigma_scenarios))
            diagonals = np.diagonal(np.dot(np.transpose(realized_sigma),realized_sigma)).copy()
    return adjusted_scenarios


def make_solution(x_dim):
    global solutions
    global sigma_matrix
    if sigma_matrix is None:
        sigma_matrix = make_sigma_matrix()
    if not solutions:
        solutions = dict()
    if not solutions.get(x_dim):
        sigma_sub_matrix = sigma_matrix[0:x_dim, 0:x_dim]
        soln = np.linalg.solve(sigma_sub_matrix,np.ones(x_dim))
        soln_sum = np.sum(soln)
        normalized_soln = [ s/soln_sum for s in soln ]
        solutions[x_dim]=normalized_soln
    return solutions[x_dim]


def markowitz_analytic_on_cube(u:[float])->float:
    """ Min-var portfolio (known solution)
    :param u:    Portfolio weights
    :return:
    """
    # Roughly equal on rays through origin
    assert all([ 0<=ui<=1 for ui in u])
    global sigma_matrix
    if sigma_matrix is None:
        sigma_matrix=make_sigma_matrix()

    x_dim = len(u)
    u_sum = np.sum(u)
    x = cube_to_weights(u)

    sigma_sub_matrix = sigma_matrix[0:x_dim,0:x_dim]
    portfolio_var = np.linalg.multi_dot( [np.array(x).transpose(), sigma_sub_matrix, x] )
    return portfolio_var+abs(1.0-u_sum)+push(u)


def push(u):
    # Tiny push away from origin
    return -0.1*min(0.1*len(u),np.sum(np.abs(u)))


def cube_to_weights(u):
    if not all([0 <= ui <= 1 for ui in u]):
        raise ValueError("u should be in hypercube")
    u_sum = np.sum(u)
    w = [(1e-8 + ui) / (1e-8 + u_sum) for ui in u]
    w = [ wi/sum(w) for wi in w ]
    return w


def realized_something_factory(g,u, adjust=False):
    """
    :param u:
    :param g:   [float] -> float  Some function on realized P/L distribution samples
    :return:
    """
    w = cube_to_weights(u)
    w_dim = len(w)

    global sigma_matrix
    global sigma_scenarios
    global adjusted_scenarios
    if sigma_matrix is None:
        sigma_matrix = make_sigma_matrix()
    if sigma_scenarios is None:
        sigma_scenarios = make_sigma_scenarios()
    if adjust and adjusted_scenarios is None:
        adjusted_scenarios = make_adjusted_scenarios()

    y = adjusted_scenarios[:, 0:w_dim] if adjust else sigma_scenarios[:, 0:w_dim]
    return g(u=u,w=w,y=np.array(y))


def markowitz_realized_on_cube(u:[float])->float:
    """ Brute force using samples, but should be the same
    :param u:
    :return:
    """
    def g(u,w,y):
        return np.var(np.dot(y, np.array(w))) + abs(1.0 - sum(u)) + push(u)

    return realized_something_factory(g=g,u=u)


def markowitz_skew_on_cube(u:[float])->float:
    """ Adds skew for fun
    :param u:
    :return:
    """
    def g(u,w,y):
        w = np.dot(y, np.array(w))
        return np.var(w) + abs(1.0 - sum(u)) + push(u) - skew(w)

    return realized_something_factory(g=g,u=u)


def markowitz_return_on_cube(u:[float])->float:
    """
    :param u:
    :return:
    """
    def g(u,w,y):
        growth = np.exp(y)
        change = (growth - 1.0)  # Could cache this
        portfolio_change = np.dot(change, np.array(w))
        log_wealth = np.log(1+portfolio_change)
        annualized_return_bp = 10000*np.mean(log_wealth)/DT
        return 0.001*abs(1.0 - sum(u)) + push(u) - annualized_return_bp

    return realized_something_factory(g=g,u=u, adjust=True)



PORTFOLIO_OBJECTIVES = [markowitz_analytic_on_cube, markowitz_realized_on_cube,
                        markowitz_return_on_cube, markowitz_skew_on_cube ]


def nice_div(a,b):
    return 0 if abs(b)<1e-6 else a/b


def troublesome_example():
    from humpday.optimizers.scipycube import scipy_nelder_cube
    from humpday.objectives.portfolio import markowitz_skew_on_cube
    v,x = scipy_nelder_cube(markowitz_skew_on_cube,n_dim=4,n_trials=100)


def markowitz_example():
    from humpday.optimizers.dlibcube import dlib_default_cube
    from humpday.optimizers.nevergradcube import nevergrad_ngopt8_cube
    from humpday.optimizers.scipycube import scipy_slsqp_cube
    from humpday.optimizers.ultraoptcube import ultraopt_gbrt_cube
    from humpday.optimizers.nloptcube import nlopt_direct_cube, nlopt_isres_cube, nlopt_esch_cube, nlopt_directr_cube
    optimizer = nlopt_directr_cube
    objective = markowitz_skew_on_cube

    n_dim = 4
    st = time.time()
    v1, u1, t1 = optimizer(objective, n_dim=n_dim, n_trials=5000, with_count=True)
    tau1 = time.time()-st

    st = time.time()
    v2, u2, t2 = optimizer(objective, n_dim=n_dim, n_trials=5000, with_count=True)
    tau2 = time.time()-st
    u3 = make_solution(x_dim=n_dim)
    v3 = objective(u3)
    v4 = markowitz_analytic_on_cube(u3)
    r = [ nice_div(u1j, u2j) for u1j, u2j in zip(u1,u2) ]
    r1 = [ nice_div(u1j, u3j) for u1j, u3j in zip(u1,u3) ]
    r2 = [ nice_div(u2j, u3j) for u2j, u3j in zip(u2,u3) ]
    results = {'r':r,'r1':r1,'r2':r2,'u1':u1,'u2':u2,'u3':u3, 'v1':v1,'v2':v2,'v3':v3, 'v4':v4,'t1':t1,'t2':t2,'tau1':tau1,'tau2':tau2}
    from pprint import pprint
    pprint(results)



if __name__=='__main__':
    troublesome_example()



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="humpday",
    version="0.6.4",
    description="Taking the pain out of choosing a Python global optimizer",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/microprediction/humpday",
    author="microprediction",
    author_email="pcotton@intechinvestments.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=["humpday", "humpday.objectives", "humpday.optimizers", "humpday.comparison", "humpday.visualization","humpday.transforms"],
    test_suite='pytest',
    tests_require=['pytest'],
    include_package_data=True,
    extras_require = {
        'full':  ["bayesian-optimization>=1.2.0","optuna","platypus-opt","poap","pysot","ultraopt","dlib",
                  "ax-platform","py-bobyqa","hebo","nlopt","freelunch"]
    },
    install_requires=["wheel","numpy>=1.21.5", "importlib-metadata>=1.7.0", "getjson",
                      "scipy", "funcy", "scipy","winning>=0.4.6","microconventions>=0.5.0",
                       "landscapes",  "pymorton","scikit-learn"],
    entry_points={
        "console_scripts": [
            "humpday=humpday.__main__:main",
        ]
    },
)



import math
import numpy as np

# http://dlib.net/optimization.html#find_min_global
# This library also provides global_function_search which is pretty darn cool


try:
    import freelunch
    using_freelunch = True
except ImportError:
    using_freelunch = False

if using_freelunch:

    def freelunch_factory(objective , method, n_trials, n_dim,  with_count=False, n_pop=None):
        global feval_count
        feval_count = 0

        def _objective(x) -> float:
            global feval_count
            feval_count += 1
            return objective(np.array(x))

        optimizer = getattr(freelunch,method)(obj=_objective, bounds=[[0, 1]] * n_dim)

        if method.lower() in ['de','sade','pso','krillherd']:
            n_gens = int(math.ceil(n_trials / n_pop))
            optimizer.hypers['N'] = n_pop   # population size
            optimizer.hypers['G'] = n_gens  # number of generations
        if method.lower() in 'sa':
            n_gens = int(math.ceil(n_trials / n_pop))
            optimizer.hypers['N'] = n_pop  # population size
            optimizer.hypers['K'] = n_gens  # number of generations

        runs = optimizer(n_runs=1, full_output=True)  # instance and run
        best_val = runs['scores'][0]  # all obj scores are sorted low to high
        best_x = runs['solutions'][0]  # corresponding inputs
        feval_count_comparison = runs['nfe']  # function evaluations

        return (best_val, best_x, feval_count) if with_count else (best_val, best_x)


    def freelunch_de_3_cube(objective, n_trials, n_dim, with_count):
        return freelunch_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 n_pop=3, method='DE')  # It is useful to have a clone of one of the better algos

    def freelunch_de_8_cube(objective ,n_trials, n_dim, with_count):
        return freelunch_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 n_pop=8, method='DE')  # It is useful to have a clone of one of the better algos


    def freelunch_de_21_cube(objective, n_trials, n_dim, with_count):
        return freelunch_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 n_pop=21, method='DE')  # It is useful to have a clone of one of the better algos


    def freelunch_sa_3_cube(objective, n_trials, n_dim, with_count):
        return freelunch_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 method='SA', n_pop=3)  # It is useful to have a clone of one of the better algos


    def freelunch_sa_8_cube(objective, n_trials, n_dim, with_count):
        return freelunch_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 method='SA', n_pop=8)  # It is useful to have a clone of one of the better algos


    def freelunch_sa_21_cube(objective, n_trials, n_dim, with_count):
        return freelunch_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 method='SA', n_pop=21)  # It is useful to have a clone of one of the better algos


    def freelunch_sade_3_cube(objective, n_trials, n_dim, with_count):
        return freelunch_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 method='SADE',n_pop=3)  # It is useful to have a clone of one of the better algos


    def freelunch_sade_8_cube(objective, n_trials, n_dim, with_count):
        return freelunch_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 method='SADE', n_pop=8)  # It is useful to have a clone of one of the better algos


    def freelunch_sade_21_cube(objective, n_trials, n_dim, with_count):
        return freelunch_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 method='SADE', n_pop=21)  # It is useful to have a clone of one of the better algos


    def freelunch_pso_3_cube(objective, n_trials, n_dim, with_count):
        return freelunch_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 method='PSO', n_pop=3)  # It is useful to have a clone of one of the better algos


    def freelunch_pso_8_cube(objective, n_trials, n_dim, with_count):
        return freelunch_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 method='PSO', n_pop=8)  # It is useful to have a clone of one of the better algos


    def freelunch_pso_21_cube(objective, n_trials, n_dim, with_count):
        return freelunch_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 method='PSO', n_pop=21)  # It is useful to have a clone of one of the better algos


    def freelunch_krillherd_3_cube(objective, n_trials, n_dim, with_count):
        return freelunch_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 method='KrillHerd', n_pop=3)  # It is useful to have a clone of one of the better algos


    def freelunch_krillherd_8_cube(objective, n_trials, n_dim, with_count):
        return freelunch_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 method='KrillHerd', n_pop=8)  # It is useful to have a clone of one of the better algos


    def freelunch_krillherd_21_cube(objective, n_trials, n_dim, with_count):
        return freelunch_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count,
                                 method='KrillHerd', n_pop=21)  # It is useful to have a clone of one of the better algos


    FREELUNCH_OPTIMIZERS = [ freelunch_sade_3_cube, freelunch_sade_8_cube, freelunch_sade_21_cube,
                             freelunch_sa_3_cube, freelunch_sa_8_cube, freelunch_sa_21_cube,
                             freelunch_de_3_cube, freelunch_de_8_cube, freelunch_de_21_cube,
                             freelunch_pso_3_cube, freelunch_pso_8_cube, freelunch_pso_21_cube,
                             freelunch_krillherd_3_cube, freelunch_krillherd_8_cube]
    FREELUNCH_TOP_OPTIMIZERS = [freelunch_sade_8_cube, freelunch_sa_21_cube]
else:
    FREELUNCH_OPTIMIZERS = []
    FREELUNCH_TOP_OPTIMIZERS = []


if __name__ == '__main__':
    from humpday.objectives.classic import CLASSIC_OBJECTIVES

    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in FREELUNCH_OPTIMIZERS:
            print(optimizer(objective, n_trials=50, n_dim=3, with_count=True))



from humpday.objectives.allobjectives import A_CLASSIC_OBJECTIVE
from humpday.optimizers.alloptimizers import OPTIMIZERS
from humpday.comparison.limitations import max_n_dim
import json
from getjson import getjson
import time
import warnings
from pprint import pprint
import traceback
import sys
import random


# Generate lookup of optimizer times. This can take a while.


def get_timing():
    return getjson('https://raw.githubusercontent.com/microprediction/humpday/main/humpday/comparison/timing.json')


def exclude(name,n_dim,n_trials):
    return 'shgo' in name and n_dim>13


def create_timing(max_elapsed=5*60):
    """ Add to the timing JSON file """

    try:
        with open('timing.json', 'rt') as fp:
            cpu = json.load(fp=fp)
    except Exception:
        cpu = dict()

    for n_dim in [2, 3, 5, 8]:
        print(' ')
        print('*********** Dimension '+str(n_dim)+' ************')
        print(' ')
        for opt in OPTIMIZERS:
            print(opt.__name__)
            dim_okay = n_dim <= max_n_dim(opt.__name__)

            if not cpu.get(opt.__name__):
                cpu[opt.__name__] = dict()
            n_failures = 0
            elapsed = -1

            if not cpu[opt.__name__].get(n_dim):
                cpu[opt.__name__][n_dim]=dict()
                for j,n_trials in enumerate([20,30,50,80,130,210,340,550]):
                    print(' '*j+str(n_trials))
                    if exclude(opt.__name__, n_dim, n_trials):
                        cpu[opt.__name__][n_dim][n_trials] = -1

                    if not cpu[opt.__name__][n_dim].get(n_trials):
                        tck = time.time()
                        try:
                            best_val,best_x = opt(objective=A_CLASSIC_OBJECTIVE,n_trials=n_trials,n_dim=n_dim)
                            elapsed = time.time()-tck
                            cpu[opt.__name__][n_dim][n_trials] = elapsed
                            if elapsed>60:
                                pprint({'name':opt.__name__,'n_dim':n_dim,'n_trials':n_trials,"elapsed":elapsed})
                        except Exception:
                            exc_type, exc_value, exc_traceback = sys.exc_info()
                            warnings.warn(opt.__name__+' failed ')
                            pprint({"name":opt.__name__,"n_dim":n_dim,"n_trials":n_trials})
                            traceback.print_tb(tb=exc_traceback)
                            cpu[opt.__name__][n_dim][n_trials] = -1
                            n_failures+=1
                        if n_failures>=1:
                            break
                        if elapsed>max_elapsed:
                            break
                    if n_failures >= 1:
                        break
                    if elapsed > 60 * 10:
                        break

            with open('timing.json','wt') as fp:
                json.dump(cpu,fp=fp)


if __name__=='__main__':
    create_timing()



from getjson import getjson
from pprint import pprint

TIMING_URL='https://raw.githubusercontent.com/microprediction/humpday/main/humpday/comparison/timing.json'


def get_timing():
    return getjson(TIMING_URL)


def get_strategy_time(optimizer,n_dim:int=5, n_trials:int=130):
    """ Return approximate time an optimizer will run for, from past trials
    :param optimizer:   str or optimizer
    :param n_dim:
    :param n_trials:
    :return:
    """
    if isinstance(optimizer,str):
        o = optimizer if '_cube' in optimizer else optimizer+'_cube'
    else:
        o = optimizer.__name__
    cpu = get_timing()
    try:
        return cpu[o][n_dim][n_trials]
    except Exception:
        return None


if __name__=='__main__':
    pprint(get_timing())


import warnings
import numpy as np

# https://github.com/numericalalgorithmsgroup/pybobyqa/blob/master/pybobyqa/solver.py

try:
    from pybobyqa import solve
    using_bobyqa = True
except ImportError:
    using_bobyqa = False

if using_bobyqa:

    def bobyqa_cube_factory(objective, n_trials, n_dim, with_count, **kwargs):
        global feval_count
        feval_count = 0

        lb = np.array([0. for _ in range(n_dim)])
        ub = np.array([1. for _ in range(n_dim)])
        x0 = np.array([0.5]*n_dim)

        def _objective(u) -> float:
            global feval_count
            feval_count += 1
            return objective(u)

        soln = solve(_objective, x0, bounds=(lb, ub), maxfun=n_trials, do_logging=False)

        best_x, best_val = list(soln.x), soln.f

        return (best_val, best_x, feval_count) if with_count else (best_val, best_x)

    def bobyqa_default_cube(objective, n_trials, n_dim, with_count):
        return bobyqa_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim, with_count=with_count)


    def bobyqa_noise_cube(objective, n_trials, n_dim, with_count):
        return bobyqa_cube_factory(objective=objective, n_trials=n_trials, n_dim=n_dim,
                                   with_count=with_count, objfun_has_noise=True)


    BOBYQA_OPTIMIZERS = [ bobyqa_default_cube, bobyqa_noise_cube ]
    BOBYQA_TOP_OPTIMIZERS = [ bobyqa_default_cube, bobyqa_noise_cube ]
else:
    BOBYQA_OPTIMIZERS = []
    BOBYQA_TOP_OPTIMIZERS = []


if __name__ == '__main__':
    from humpday.objectives.classic import CLASSIC_OBJECTIVES

    for objective in CLASSIC_OBJECTIVES:
        print(' ')
        print(objective.__name__)
        for optimizer in BOBYQA_OPTIMIZERS:
            print(optimizer(objective, n_trials=50, n_dim=34, with_count=True))